/******************************************************************************/
/* rund.cc                 : Test MySQL multi thread.                         */
/* By                      : Edwin Vandamme                                   */
/* Creation                : 03/10/2011                                       */
/* Version                 : 1.0.0                                            */
/* Modifications           :                                                  */
/*                                                                            */
/******************************************************************************/
/* REMARKS :                                                                  */
/*                                                                            */
/******************************************************************************/

// -- Libraries. --------------------------------------------------------------/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>                                                            // Shared memory.
#include <fcntl.h>                                                              // File control.
#include <errno.h>                                                              // Error handling.
#include <pwd.h>
#include <iostream>
#include <fstream>                                                              // Required to read files.
#include <stdarg.h>
#include <mysql/mysql.h>                                                        // MySQL library.
#include <mysql/errmsg.h>                                                       // MySQL error messages.
#include <pthread.h>                                                            // pThread library.
#include <vector>                                                               // Dynamic array library.
#include <wait.h>                                                               // Required for external program call.
#include <arpa/inet.h>                                                          // Required for socket connections.
#include <netdb.h>                                                              // Required for socket connections.
#include <netinet/in.h>                                                         // Required for socket connections.
#include <signal.h>                                                             // Required to handle signals (e.g. eliminate zombie processes.)
#include <math.h>                                                               // Mathematical functions.

//#include "/Programs/includes/isdn.h"                                            // Switch constants and structures.

#define LOCKMODE (S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH)                              // Lock file mode.

#define lLogM    200                                                            // Width of a log message.
#define DAEMON_NAME "rund"                                                      // Daemon runs as ...


using namespace std;

// -- Global structures. ------------------------------------------------------/
// +1 Is used in char arrays to allow for the NULL char at the end of a string.
struct GLOBALS_t
   {
   int              PID;                                                        // PID.
   pthread_mutex_t  logF_m;                                                     // Lock/unlock log processing.
//   FILE             *logF;                                                      // Handle for log file.
   long             logFS;                                                      // Max. log file size before log rotation.
   string           logFN;                                                      // Log file name and location.
   int              IServer;                                                    // switchd server id.
   int              IUnit;                                                      // switchd unit id.
   string           daemon_u;                                                   // switchd runs as user ...
   string           daemon_n;                                                   // Daemon name "switchd".
   vector <string>  mysql_h;                                                    // Array with possible MySQL hosts.
   string           mysql_u;                                                    // MySQL connection user name.
   string           mysql_pw;                                                   // MySQL connection password.
   string           mysql_n;                                                    // MySQL cluster check table.
   int              mysql_p;                                                    // MySQL connection port.
   pthread_mutex_t  mysql_m;                                                    // Only 1 process uses the mysql connection at anny given time.
//   pthread_mutex_t  mysql_x;                                                    // Only 1 process uses the mysql connection at anny given time (extra safety lock).
   MYSQL            *mysql;                                                     // MySQL connection.
   int              nISDN;                                                      // Number of ISDN lines configured.
   bool             log_x;                                                      // Log progress in switch functions.
   bool             log_q;                                                      // Log MySQL queries.
   bool             log_m;                                                      // Log MySQL information.
   };

struct ISDN_t                                                                   // ISDN line structure.
   {
   int              Id;                                                         // ISDN line number.
   pthread_t        tId;                                                        // Thread id.
   int              tRet;                                                       // Thread return.
   char             logM      [lLogM+1];                                        // Error message buffer specific for each line, to avoid thread conflicts.
   pthread_mutex_t  m;                                                          // Lock/unlock line processing.
   pthread_cond_t   c;                                                          // Wait condition for line processing.
   };

struct shmISDN_t                                                                // ISDN lines shared memory.
   {
   int              shmId;                                                      // Shared memory id.
   ISDN_t           *shmAddr;                                                   // Shared memory address.
   };
   
// -- Global settings. --------------------------------------------------------/
int             progMax         = 100;                                          // Max. times progTimer may elapse before a program is terminated.
int             progTimer       = 100000;                                       // Time between program status checks (micro seconds 1sec=1000000usec).
//int             mQuery          = 10000;                                        // Max. number of queries to execute before reconnecting to MySQL to regain memory.

// -- General settings. -------------------------------------------------------/
int             dummy           = 0;
bool            daemon_o        = true;                                         // switchd On indicator.
string          daemon_h        = "";                                           // switchd hostname.
string          daemon_i        = "";                                           // switchd ip address.
size_t          daemon_hL;                                                      // switchd hostname length.
pthread_attr_t  daemon_attr;                                                    // switchd general thread settings.

// -- Ticker, timed events settings. -------------------------------------------/
pthread_t       ticker_t;                                                       // Ticker, timed events, thread id.
int             ticker_r;                                                       // Ticker, timed events, thread return.
int             ticker_base     = 10000000;                                     // Ticker base event every ... milliseconds. (1000000 = 1 sec)
int             ticker_mysql_e  = 1;                                            // Ticker base event multiplier for mysql connection hartbeat.
int             ticker_mysql    = 0;                                            // Ticker base event counter for mysql connection hartbeat.

// -- Dialogic settings. ------------------------------------------------------/
shmISDN_t       *lISDN;                                                         // Pointer to ISDN line array.
pthread_mutex_t ISDN_m          = PTHREAD_MUTEX_INITIALIZER;                    // All ISDN lines wait for a single ISDN line to complete.

// -- Global variables. -------------------------------------------------------/
GLOBALS_t       g;                                                              // Global variables, used in switchd and functions.
int             lSocket;                                                        // Listening socket.
int             shmSize;                                                        // Shared memory segment size.
key_t           shmKey;                                                         // Shared memory key.
//int             nQuery          = 0;                                            // Query count.

// -- Function declarations. --------------------------------------------------/
void clean_shutDown (int ExitCode);
static void child_handler (int signum);
static void daemonize (const char *lock_file);
void *handleISDN (void *ptr);
void *handleTicker (void *ptr);

// Initialize global variables.
void initGlobals (GLOBALS_t &g)
   {
   pthread_mutex_init (&g.logF_m, NULL);
   pthread_mutex_init (&g.mysql_m, NULL);
//   pthread_mutex_init (&g.mysql_x, NULL);
   g.PID             = 0;
//   g.logF            = NULL;
   g.logFN           = "/Programs/functions/test.mysql/rund.log";
   g.logFS           = 50000000;
   g.IServer         = 0;
   g.IUnit           = 0;
   g.daemon_u        = "root";
   g.daemon_n        = "rund";
   g.mysql_h.push_back ("localhost");
   g.mysql_u         = "root"; 
   g.mysql_pw        = "";
   g.mysql_n         = "qualitycam.ndb_check";
   g.mysql_p         = 3306;

//   g.nISDN           = 5;
   g.nISDN           = 3;
   g.log_x           = true;
   g.log_q           = false;
   g.log_m           = true;

   }

// Get memory usage.
unsigned long getmem (int PID)
   {
   struct proc_t {
      int               pid;           // process id
      char              cmd[16];       // basename of executable file in call to exec(2)
      char              state;         // single-char code for process state (S=sleeping)
      int               ppid;          // pid of parent process
      int               pgrp;          // process group id
      int               session;       // session id
      int               tty;           // full device number of controlling terminal
      int               tpgid;         // terminal process group id
      unsigned long     flags;         // kernel flags for the process
      unsigned long     min_flt;       // number of minor page faults since process start
      unsigned long     cmin_flt;      // cumulative min_flt of process and child processes
      unsigned long     maj_flt;       // number of major page faults since process start
      unsigned long     cmaj_flt;      // cumulative maj_flt of process and child processes
      unsigned long     utime;         // user-mode CPU time accumulated by process
      unsigned long     stime;         // kernel-mode CPU time accumulated by process
      long              cutime;        // cumulative utime of process and reaped children
      long              cstime;        // cumulative stime of process and reaped children
      long              priority;      // kernel scheduling priority
      long              nice;          // standard unix nice level of process
      long              timeout;       // ?
      long              it_real_value; // ?
      unsigned long     start_time;    // start time of process -- seconds since 1-1-70
      unsigned long     vsize;         // number of pages of virtual memory ...
      long              rss;           // resident set size from /proc/#/stat
      unsigned long     rss_rlim;      // resident set size ... ?
      unsigned long     start_code;    // address of beginning of code segment
      unsigned long     end_code;      // address of end of code segment
      unsigned long     start_stack;   // address of the bottom of stack for the process
      unsigned long     kstk_esp;      // kernel stack pointer
      unsigned long     kstk_eip;      // kernel stack pointer
      unsigned long     wchan;         // address of kernel wait channel proc is sleeping in
      unsigned long     nswap;         // ?
      unsigned long     cnswap;        // cumulative nswap ?
      int               lproc;         // last processor
   };

   FILE       *memF;
   char        L [500];
   char        pidFN [500];
   proc_t      P;

   sprintf (pidFN, "/proc/%i/stat", PID);
   memF = fopen (pidFN, "r");
   if (memF != NULL)
      {
      if (fgets (L, sizeof (L), memF) != NULL)
         {
         char* tmp = strrchr (L, ')');                                          // split into "PID (cmd" and "<rest>"
         *tmp = '\0';                                                           // replace trailing ')' with NUL

         // Parse these two strings separately, skipping the leading "(".
         memset (P.cmd, 0, sizeof (P.cmd));                                     // clear even though *P xcalloc'd ?!
         sscanf (L, "%d (%15c", &P.pid, P.cmd);                                 // comm[16] in kernel

         sscanf (tmp+2, "%c "
                        "%d %d %d %d %d "
                        "%lu %lu %lu %lu %lu %lu %lu "
                        "%ld %ld %ld %ld %ld %ld "
                        "%lu %lu "
                        "%ld "
                        "%lu %lu %lu %lu %lu %lu "
                        "%*s %*s %*s %*s " 
                        "%lu %lu %lu %*d %d",
                        &P.state,
                              &P.ppid, &P.pgrp, &P.session, &P.tty, &P.tpgid,
                        &P.flags, &P.min_flt, &P.cmin_flt, &P.maj_flt, &P.cmaj_flt, &P.utime, &P.stime,
                        &P.cutime, &P.cstime, &P.priority, &P.nice, &P.timeout, &P.it_real_value,
                        &P.start_time, &P.vsize,
                        &P.rss,
                        &P.rss_rlim, &P.start_code, &P.end_code, &P.start_stack, &P.kstk_esp, &P.kstk_eip,
                        &P.wchan, &P.nswap, &P.cnswap, &P.lproc);
         }
      fclose (memF);
      }

   return P.vsize;
   }

// Check if a file exists.
bool fexists (const char *filename)
   {
   ifstream ifile (filename);
   return ifile;
   }

// Check a file lock.
int lockfile (int fd)
   {
   struct flock fl;

   fl.l_type   = F_WRLCK;
   fl.l_start  = 0;
   fl.l_whence = SEEK_SET;
   fl.l_len    = 0;

   return (fcntl (fd, F_SETLK, &fl));
   }

// Copy a string in a fixed length string, make sure len is set correctly.
void fstrncpy (char * t, char * f, int len)
   {
   strncpy (t, f, (len-1));
   t [(len-1)] = '\0';
   }

// Write to the log file.
void writeLog (GLOBALS_t &g, char * M, bool OnScreen=false, bool dt=true)
   {
   pthread_mutex_lock (&g.logF_m);
   time_t     now;
//   struct tm  *ti;
   struct tm   ti;
   FILE       *logF;
   long        i;
   string      S;
   char        T [20];

   
   logF = fopen (g.logFN.c_str (), "a+");
   if (logF != NULL)
      {
      fseek (logF, 0, SEEK_END);
      i = ftell (logF);
      fclose (logF);

      if (i > g.logFS)
         {
         i = 0;
         S = g.logFN;
         while (fexists (S.c_str ()))
            {
            i++;
            if (i > 999) {break;}
            S = g.logFN;
            sprintf (T, ".%03li", i);
            S.append (T, sizeof (T));
            }
         if (i <=  999) {rename (g.logFN.c_str (), S.c_str ());}
         }
      }
      
   logF = fopen (g.logFN.c_str (), "a+");
   if (logF == NULL)
      {
      printf ("Unable to open %s, errno = 0x%X.\n", g.logFN.c_str (), errno);
      }
     else
      {
      if (dt)
         {
         time (&now);
//         ti = localtime (&now);
         localtime_r (&now, &ti);
//         if (g.PID > 0)
//            {if (logF!=NULL) {fprintf (logF, "[%04i/%02i/%02i %02i:%02i:%02i] %s : %15lu : %s", (ti->tm_year+1900), (ti->tm_mon+1), ti->tm_mday, ti->tm_hour, ti->tm_min, ti->tm_sec, g.daemon_n.c_str (), getmem (g.PID), M);}}
//           else
//            {if (logF!=NULL) {fprintf (logF, "[%04i/%02i/%02i %02i:%02i:%02i] %s : %15i : %s", (ti->tm_year+1900), (ti->tm_mon+1), ti->tm_mday, ti->tm_hour, ti->tm_min, ti->tm_sec, g.daemon_n.c_str (), 0, M);}}
         if (g.PID > 0)
            {if (logF!=NULL) {fprintf (logF, "[%04i/%02i/%02i %02i:%02i:%02i] %s : %15lu : %s", (ti.tm_year+1900), (ti.tm_mon+1), ti.tm_mday, ti.tm_hour, ti.tm_min, ti.tm_sec, g.daemon_n.c_str (), getmem (g.PID), M);}}
           else
            {if (logF!=NULL) {fprintf (logF, "[%04i/%02i/%02i %02i:%02i:%02i] %s : %15i : %s", (ti.tm_year+1900), (ti.tm_mon+1), ti.tm_mday, ti.tm_hour, ti.tm_min, ti.tm_sec, g.daemon_n.c_str (), 0, M);}}
//         if (logF!=NULL) {fprintf (logF, "[2011/10/28 01:02:03] %s : %15lu : %s", g.daemon_n.c_str (), getmem (g.PID), M);}
         }
        else
         {if (logF!=NULL) {fprintf (logF, "%s", M);}}
      if (OnScreen) {printf (M);}
      if (logF!=NULL) {fflush (logF);}
      fclose (logF);
      }
   pthread_mutex_unlock (&g.logF_m);
   }

// Shutdown application.
void shutDown (GLOBALS_t &g, int ExitCode)
   {
   char            M [200];                                                     // Message buffer.

   // Close MySQL connection.
   if (g.log_x) {sprintf (M, "==>   : 000 : shutDown () : close MySQL\n"); writeLog (g, M);}
   mysql_close (g.mysql);
   if (g.log_x) {sprintf (M, "===   : 000 : shutDown () : close MySQL mysql_close\n"); writeLog (g, M);}
   mysql_server_end ();
   if (g.log_x) {sprintf (M, "<==   : 000 : shutDown () : close MySQL mysql_server_end\n"); writeLog (g, M);}

   exit (ExitCode);
   }

void clean_shutDown (int ExitCode)
   {
   char            M [200];                                                     // Message buffer.

   if (g.log_x) {sprintf (M, "==>   : 000 : clean_shutDown () : \n"); writeLog (g, M);}

   // Stop running threads at end of program.
   daemon_o = false;
   if (g.log_x) {sprintf (M, "==>   : 000 : stopping threads.\n"); writeLog (g, M);}
   for (int i=1; i<=g.nISDN; i++)
      {
      pthread_cond_broadcast (&(*lISDN [i-1].shmAddr).c);
      }
   usleep (ticker_base+5000000);
   if (g.log_x) {sprintf (M, "<==   : 000 : stopping threads.\n"); writeLog (g, M);}
   
   // Remove shared memory segments.
   // If not a reboot is required to free the memory before a bigger segment can be allocated.
   for (int i=1; i<=g.nISDN; i++)
      {
      shmdt (lISDN [i-1].shmAddr);
      shmctl (lISDN [i-1].shmId, IPC_RMID, NULL);
      }
   delete [] lISDN;

   // Finish up.
   sprintf (M, "Info  : Stopped %s on %s.\n", DAEMON_NAME, daemon_h.c_str ());
   writeLog (g, M);

   if (g.log_x) {sprintf (M, "<==   : 000 : clean_shutDown () : \n"); writeLog (g, M);}

   shutDown (g, ExitCode);
   }

// Set up MySQL connection.
MYSQL *connectMySQL (GLOBALS_t &g)
   {
   char            M [200];                                                     // Message buffer.
   MYSQL           *mysql;                                                      // MySQL connection.
   MYSQL_RES       *results;                                                    // MySQL returned results.
   char            Query [100];                                                 // MySQL query.
   int             mysql_e;                                                     // MySQL error.

   mysql = mysql_init (NULL);
   mysql_options (mysql, MYSQL_READ_DEFAULT_GROUP, "libmysqld_client");
   mysql_options (mysql, MYSQL_OPT_USE_EMBEDDED_CONNECTION, NULL);
   for (int i=1;i<=(int) g.mysql_h.size ();i++)
      {
      mysql_e = 0;
      if (!mysql_real_connect (mysql, g.mysql_h [i-1].c_str (), g.mysql_u.c_str(), g.mysql_pw.c_str(), "", g.mysql_p, NULL, 0))
         {sprintf (M, "Error : MySQL : unable to connect to MySQL server %i/%i, error : %s.\n", i, g.mysql_h.size (), mysql_error (mysql)); writeLog (g, M, false, true);}
        else
         {
         // Check cluster connection.
         if (g.mysql_n != "")
            {
            sprintf (Query, "SELECT * FROM %s LIMIT 1;", g.mysql_n.c_str ());
            mysql_query (mysql, Query);
            results = mysql_store_result (mysql);
            mysql_e = mysql_errno (mysql);
            if (mysql_e != 0) {sprintf (M, "Error : MySQL : NDB server %i/%i : failed %s\n", i, g.mysql_h.size (), mysql_error (mysql)); writeLog (g, M, false, true);}
            if (results) {mysql_free_result (results);}
            }
            
         if (mysql_e == 0) {if (g.log_m) {sprintf (M, "Info  : MySQL : connected to MySQL server %i/%i on %s.\n", i, g.mysql_h.size (), g.mysql_h [i-1].c_str ()); writeLog (g, M, false, true);} break;}
         }

      if (i == (int) g.mysql_h.size ())
         {
         sprintf (M, "Error : MySQL : unable to connect to anny MySQL server.\n");
         writeLog (g, M, false, true);
         shutDown (g, EXIT_FAILURE);
         }
      }
      
   return mysql;
   }

// Trim trailing characters.
string rtrim (string S, string C = " \t")
   {
   size_t E = S.find_last_not_of (C);
   if (string::npos != E) {S = S.substr (0, E+1);}
   return S;
   }

// Execute a query and recover upon failure.
void doQ (int iISDN, GLOBALS_t &g, MYSQL_RES *& results, char *Query)
   {
   pthread_mutex_lock (&g.mysql_m);
   char            M [2000];                                                    // Message buffer.
   int             E;
   bool            bResult;

/*
   string          S;
   S = Query;
   sprintf (M, "Info  : MySQL : %s\n", S.c_str ()); writeLog (g, M, false, true);
   fstrncpy (Query, (char *) rtrim (S, ";").c_str (), sizeof (Query));
   sprintf (M, "Info  : MySQL : %s\n", Query); writeLog (g, M, false, true);
*/
/*
   // Close/open MySQL connection to try to avoid memory bug.
   nQuery++;
   if (nQuery > mQuery)
      {
      nQuery = 0;
      if (g.log_x) {sprintf (M, "==>   : %03i : CLOSE MySQL ()\n", iISDN); writeLog (g, M);}
      // Close MySQL connection.
      mysql_close (g.mysql);
      if (g.log_x) {sprintf (M, "<==   : %03i : CLOSE MySQL ()\n", iISDN); writeLog (g, M);}
// Closing the library takes 5 seconds.
//      mysql_library_end ();
//      mysql_server_end ();
//      if (g.log_x) {sprintf (M, "<==   : %03i : END   MySQL ()\n", iISDN); writeLog (g, M);}
      usleep (500000);
      g.mysql = connectMySQL (g);
      if (g.log_x) {sprintf (M, "<==   : %03i : OPEN  MySQL ()\n", iISDN); writeLog (g, M);}
      }
*/
      
   if (g.log_x) {sprintf (M, "==>   : %03i : doQ ()\n", iISDN); writeLog (g, M);}
/**/
   if (g.log_q)
      {
      sprintf (M, "Info  : MySQL : %s \n", Query);
      writeLog (g, M, false, true);
      }

   // Check if a result is expected to be returned.
   fstrncpy (M, Query, 7);
//   if (results) {mysql_free_result (results);} else {results = NULL;}
   if (strcmp (M, "SELECT") == 0) {bResult = true;} else {bResult = false;}

   // Execute query.
sprintf (M, "===   : %03i : doQ () - 001\n", iISDN); writeLog (g, M, false, true);
/**/
//   if (mysql_real_query (g.mysql, Query, strlen (Query)) != 0)
//   pthread_mutex_lock (&g.mysql_x);
   if (mysql_query (g.mysql, Query) != 0)
      {
sprintf (M, "===   : %03i : doQ () - 002\n", iISDN); writeLog (g, M, false, true);
      E = mysql_errno (g.mysql);
      sprintf (M, "Error : MySQL : %i : (1) %s\n", E, mysql_error (g.mysql));
      writeLog (g, M, false, true);
      sprintf (M, "Error : MySQL : %i : (1) unable to execute query : %s\n", E, Query);
      writeLog (g, M, false, true);

      // Reconnect to MySQL server.
      if ((E == CR_SERVER_GONE_ERROR) || (E == CR_SERVER_LOST)) {g.mysql = connectMySQL (g);}

      // Retry query.
      if (mysql_query (g.mysql, Query) != 0)
         {
sprintf (M, "===   : %03i : doQ () - 003\n", iISDN); writeLog (g, M, false, true);
         E = mysql_errno (g.mysql);
         sprintf (M, "Error : MySQL : %i : (2) %s\n", E, mysql_error (g.mysql));
         writeLog (g, M, false, true);
         sprintf (M, "Error : MySQL : %i : (2) failed to execute query : %s\n", E, Query);
         writeLog (g, M, false, true);
         }
      }
   // Obtain query results.
sprintf (M, "===   : %03i : doQ () - 004\n", iISDN); writeLog (g, M, false, true);
   if (bResult) {results = mysql_store_result (g.mysql);}
//   pthread_mutex_unlock (&g.mysql_x);
/**/
/*
// micro seconds 1sec=1000000usec Query takes 0.007 (0.0284) secs - 7000/28400
E = rand () % 10000 + 20000;
sprintf (M, "===   : %03i : doQ (%i) - 004\n", iISDN, E); writeLog (g, M, false, true);
*/
   if (g.log_x) {sprintf (M, "<==   : %03i : doQ ()\n", iISDN); writeLog (g, M);}
   pthread_mutex_unlock (&g.mysql_m);
//usleep (E);
   }

static void child_handler (int signum)
   {
   switch (signum)
      {
      case SIGALRM: clean_shutDown (EXIT_FAILURE); break;
      case SIGUSR1: clean_shutDown (EXIT_SUCCESS); break;
      case SIGUSR2: exit (EXIT_SUCCESS);           break;                       // Shutdown of daemonizing process.
      case SIGCHLD: clean_shutDown (EXIT_FAILURE); break;
      }
   }

static void daemonize (const char *lock_file)
   {
   char            M [200];                                                     // Message buffer.
   pid_t           pid, sid, parent;
   int             lfp = -1;
   char            pidbuf [16];

   // Already a daemon.
   if (getppid () == 1) return;

   // Drop user if there is one, and we were run as root.
   if (getuid () == 0 || geteuid () == 0)
       {
       struct passwd *pw = getpwnam (g.daemon_u.c_str ());
       if (pw) {setuid (pw->pw_uid);}
       }

   // Trap signals that we expect to recieve.
   signal (SIGCHLD, child_handler);
   signal (SIGUSR1, child_handler);
   signal (SIGUSR2, child_handler);
   signal (SIGALRM, child_handler);
   signal (SIGHUP,  child_handler);

   // Fork off the parent process.
   pid = fork ();
   if (pid < 0)
      {
      sprintf (M, "Error : unable to fork daemon, code=%d (%s)\n", errno, strerror (errno));
      writeLog (g, M, true);
      shutDown (g, EXIT_FAILURE);
      }

   // If we got a good PID, then we can exit the parent process.
   if (pid > 0)
      {
      // Stop parent process.
      exit (EXIT_SUCCESS);
      }

   // At this point we are executing as the child process.
   parent = getppid ();

   // Create the lock file as the current user.
   if (lock_file && lock_file [0])
      {
      lfp = open (lock_file, O_RDWR|O_CREAT, LOCKMODE);
      if (lfp < 0)
         {
         sprintf (M, "Error : unable to open lock %s, code=%d (%s)\n", lock_file, errno, strerror (errno));
         writeLog (g, M, true);
         kill (parent, SIGALRM);
         shutDown (g, EXIT_FAILURE);
         }
      if (lockfile (lfp) < 0)
         {
         if (errno == EACCES || errno == EAGAIN)
            {
            sprintf (M, "Info  : locked %s\n", lock_file);
            writeLog (g, M);
            kill (parent, SIGALRM);
            shutDown (g, EXIT_FAILURE);
            }
           else
            {
            sprintf (M, "Error : unable to lock %s, code=%d (%s)\n", lock_file, errno, strerror (errno));
            writeLog (g, M, true);
            kill (parent, SIGALRM);
            shutDown (g, EXIT_FAILURE);
            }
         }
      ftruncate (lfp, 0);
      sprintf (pidbuf, "%ld", (long) getpid ());
      write (lfp, pidbuf, strlen (pidbuf)+1);
      }

   // Cancel certain signals.
   signal (SIGCHLD, SIG_DFL);                                                  // A child process dies.
   signal (SIGTSTP, SIG_IGN);                                                  // Various TTY signals.
   signal (SIGTTOU, SIG_IGN);
   signal (SIGTTIN, SIG_IGN);
   signal (SIGTERM, SIG_DFL);                                                  // Die on SIGTERM.

   // Change the file mode mask.
   umask (0);

   // Create a new SID for the child process.
   sid = setsid ();
   if (sid < 0)
      {
      sprintf (M, "Error : unable to create a new session, code %d (%s)\n", errno, strerror (errno));
      writeLog (g, M, true);
      shutDown (g, EXIT_FAILURE);
      }
   g.PID = sid;

   // Change the current working directory.  This prevents the current directory from being locked; hence not being able to remove it.
   if ((chdir ("/")) < 0)
      {
      sprintf (M, "Error : unable to change directory to %s, code %d (%s)\n", "/", errno, strerror (errno));
      writeLog (g, M, true);
      shutDown (g, EXIT_FAILURE);
      }

   // Redirect standard files to /dev/null.
   freopen ("/dev/null", "r", stdin);
// ===> Enable after everything is done ? solved. (when enabled, PuTTY connections are propperly cleared.)
   freopen ("/dev/null", "w", stdout);
   freopen ("/dev/null", "w", stderr);

   sprintf (M, "Info  : PID : %i,%i,%i\n", parent, sid, pid); writeLog (g, M);

   // Tell the parent process that we are A-okay.
   kill (parent, SIGUSR2);
   }

// Ticker, timed events.
void *handleTicker (void *ptr)
   {
   char            M [200];                                                     // Message buffer.
   int             rc;

   sprintf (M, "Info  : starting timed events.\n");
   writeLog (g, M);
   mysql_thread_init ();
   while (daemon_o)
      {
      // Update tickers.
      ticker_mysql++;
      
      // Keep mysql_connection alive by pinging regulary.
      if (ticker_mysql >= ticker_mysql_e)
         {
         ticker_mysql = 0;
         pthread_mutex_lock (&g.mysql_m);
//         pthread_mutex_lock (&g.mysql_x);
         if (mysql_ping (g.mysql) > 0)
            {
            sprintf (M, "Error : MySQL connection lost, error : %s.\n", mysql_error (g.mysql));
            writeLog (g, M);
            g.mysql = connectMySQL (g);
            }
//         pthread_mutex_unlock (&g.mysql_x);
         pthread_mutex_unlock (&g.mysql_m);
         }

      usleep (ticker_base);
      }

   mysql_thread_end ();
   if ((rc = pthread_detach (ticker_t)) != 0) {sprintf (M, "Error : 000 : Unable to detach ticker thread : %i.\n", rc); writeLog (g, M);}
   pthread_exit (NULL);
   }

// Handle ISDN lines.
void *handleISDN (void *ptr)
   {
   char            M [200];                                                     // Message buffer.
   char            Query [1000];                                                // MySQL query.
   MYSQL_RES       *results = NULL;                                             // MySQL result set.
   MYSQL_ROW       record;                                                      // MySQL record buffer.
   int             iISDN;                                                       // Line id.
   string          S = "";                                                      // Temporary string.
   int             rc;

   iISDN = (int) ptr;
   mysql_thread_init ();

   while (daemon_o)
      {
      // Test MySQL SELECT.
      sprintf (Query, "SELECT IServer,Unit,Script,Node,Type,OnKey1,OnKey2,OnKey3,OnKey4,OnKey5,OnKey6,OnKey7,OnKey8,OnKey9,OnKey0,OnKeyP,OnKeyA,ClearKeys,WaitMess,Digits,TimeOut,OnTimeOut,Tel,Message,Function FROM callsystem.scripts WHERE ((IServer=0 AND Unit=0 AND Script='HIJGEN') OR (IServer=1 AND Unit=0 AND Script='')) AND Node='2002' ORDER BY IServer, Script, Node LIMIT 1;");
//      sprintf (Query, "SELECT IServer,Unit,Script,Node,Type,OnKey1,OnKey2,OnKey3,OnKey4,OnKey5,OnKey6,OnKey7,OnKey8,OnKey9,OnKey0,OnKeyP,OnKeyA,ClearKeys,WaitMess,Digits,TimeOut,OnTimeOut,Tel,Message,Function FROM callsystem.scripts WHERE ((IServer=0 AND Unit=0 AND Script='HIJGEN') OR (IServer=1 AND Unit=0 AND Script='')) AND Node='2002' ORDER BY IServer, Script, Node LIMIT 1");
      doQ (iISDN, g, results, Query);
      if (results && (record = mysql_fetch_row (results)))
         {
//sprintf (M, "Result:%s,%s,%s,%s\n", record [0], record [1], record [2], record [3]); writeLog (g, M);
//         usleep (100*iISDN);
         usleep (100);
         }
//      if (results) {mysql_free_result (results);}
      if (results) {mysql_free_result (results);} else {results = NULL;}
      }

   mysql_thread_end ();
   if ((rc = pthread_detach ((*lISDN [iISDN-1].shmAddr).tId)) != 0) {sprintf (M, "Error : %03i : Unable to detach line thread : %i.\n", iISDN, rc); writeLog (g, M);}
   pthread_exit (NULL);
   }

int main (int argc, char *argv [])
   {
   char            M [200];                                                     // Message buffer.
   char            Query [1000];                                                // MySQL query.
   size_t          thread_stacksize;


   // Initialize globals.
   initGlobals (g);
   srand (time (NULL));

   pthread_attr_getstacksize (&daemon_attr, &thread_stacksize);
   sprintf (M, "Info  : Thread default stack size %li.\n", (long int)thread_stacksize); writeLog (g, M);

   // Initialize stack size to 256Kb.
   pthread_attr_init (&daemon_attr);
   pthread_attr_setstacksize (&daemon_attr, (256*1024));
   
   // Segment id, based upon a folder.
   shmKey  = ftok ("./", 'a');
   shmSize = sizeof (ISDN_t);

   // Initialize global values.
   daemon_hL = sizeof (Query);
   gethostname (Query, daemon_hL);
   daemon_h = Query;
   if (daemon_h == "") {daemon_h = DAEMON_NAME;}
   
   // Daemonize.
   daemonize ("/var/lock/subsys/" DAEMON_NAME);

   // Avoid zombie processes.
   signal (SIGCHLD, SIG_IGN);

   // Log start.
   sprintf (M, "Info  : Starting %s on %s.\n", DAEMON_NAME, daemon_h.c_str ());
   writeLog (g, M);

   // MySQL initializations.
   g.mysql = connectMySQL (g);

   // Initialize ISDN lines.
   lISDN    = new shmISDN_t [g.nISDN];

   // Try to open shared memory segments and close if they exist.
   for (int i=1; i<=g.nISDN; i++) {if ((lISDN [i-1].shmId = shmget ((shmKey+i), sizeof ((*lISDN)), 0666)) >= 0) {shmctl (lISDN [i-1].shmId, IPC_RMID, NULL);}}

   // Initialize ISDN structures.
   for (int i=1; i<=g.nISDN; i++)
      {
      // Create memory segment for each line.
      if ((lISDN [i-1].shmId = shmget ((shmKey+i), shmSize, IPC_CREAT | 0666)) < 0)
         {
         sprintf (M, "Error : unable to create shared memory segment for a line (Line:%i, %i bytes).\n", i, shmSize);
         writeLog (g, M, true);
         shutDown (g, EXIT_FAILURE);
         }

      // Attach segment.
      if ((lISDN [i-1].shmAddr = (ISDN_t *) shmat (lISDN [i-1].shmId, NULL, 0)) == (ISDN_t *)-1)
         {
         sprintf (M, "Error : unable to attach shared memory segment for a line.\n");
         writeLog (g, M, true);
         shutDown (g, EXIT_FAILURE);
         }

      // Set ISDN_t data.
      (*lISDN [i-1].shmAddr).Id      = i;
      }

   // Start ISDN threads.
   for (int i=1; i<=g.nISDN; i++)
      {
      sprintf (M, "Info  : Starting Line %i.\n", i);
      writeLog (g, M);

      // Set initial line data.
      pthread_mutex_init (&(*lISDN [i-1].shmAddr).m, NULL);
      pthread_cond_init (&(*lISDN [i-1].shmAddr).c, NULL);
      (*lISDN [i-1].shmAddr).tRet = pthread_create (&(*lISDN [i-1].shmAddr).tId, &daemon_attr, handleISDN, (void*) (*lISDN [i-1].shmAddr).Id);
      }

   // Start timed events = ticker thread.
   ticker_r = pthread_create (&ticker_t, &daemon_attr, handleTicker, (void*) dummy);

   // Wait for all threads to end before continuing.
   for (int i=1; i<=g.nISDN; i++)
      {
      pthread_join ((*lISDN [i-1].shmAddr).tId, NULL); usleep (1);
      }
   pthread_join (ticker_t, NULL); usleep (1);

//   clean_shutDown (EXIT_SUCCESS);
   return 0;
   }
